service = """
apiVersion: v1
kind: Service
metadata:
  name: service-easier-serving
  labels:
    app: easier-serving
spec:
  type: NodePort
  # type: LoadBalancer
  ports:
  - name: "5000"
    port: 5000
    targetPort: 5000
  selector:
    app: easier-serving
"""


#   Copyright  2020 Atos Spain SA. All rights reserved.
 
#   This file is part of EASIER AI.
 
#   EASIER AI is free software: you can redistribute it and/or modify it under the terms of Apache License, either version 2 of the License, or
#   (at your option) any later version.
 
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT ANY WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING 
#   BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT,
#   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#   WHETHER IN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
#   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#   See  LICENSE file for full license information  in the project root.

import sys
sys.path.append('..')
sys.path.append('.')

import time
import json

from easierSDK.classes.categories import Categories

class ModelMetadata():
    """Class to store the model's metadata information.
    """

    category = ''
    name = ''
    last_modified = ''
    description = ''
    version = 0
    features = []
    previous_experimentID = -1
    experimentIDs = []

    def __init__(self, f:dict=None):
        """Constructor for the EasierModel Metadata Class.

        Args:
            f (dict, optional): Dictionary with the model's metadata information.
        """
        if f is not None :
            if 'category' in f: 
                try:
                    self.category = Categories(f['category'])
                except:
                    print("[ModelMetadata] Category " + str(f['category']) + " not recognized. Using 'misc' as categoriy")
                    self.category = Categories.MISC
            if 'name' in f: self.name = f['name']
            if 'last_modified' in f: 
                try:
                    self.last_modified = time.strftime('%H:%M:%S - %d/%m/%Y', f['last_modified'])
                except:
                    # print("[ModelMetadata] Couldn't parse time in 'last-modified'. Keeping as str")
                    self.last_modified = f['last_modified']
            if 'description' in f: self.description = f['description']
            if 'version' in f: self.version = f['version']
            if 'features' in f: self.features = f['features']
            if 'previous_experimentID' in f: self.previous_experimentID = int(f['previous_experimentID'])
        else:
            pass

    # def seria
    def pretty_print(self):
        """Print the metadata information.
        """
        if isinstance(self.category, str):
            category = Categories[category.upper()]
        else:
            category = self.category.value
        row_format ="{:<30}" * 2
        print(row_format.format(*['Category:', category]))
        print(row_format.format(*['Name:', self.name]))
        print(row_format.format(*['Description:', self.description]))
        print(row_format.format(*['Last modified:', str(self.last_modified)]))
        print(row_format.format(*['Version:', self.version]))
        print(row_format.format(*['Features:', str(self.features)]))
        if int(self.previous_experimentID) > -1: print(row_format.format(*['previous_experimentID:', str(self.previous_experimentID)]))

    def dump_to_file(self, path:str):
        """Dump metadata information to a file.

        Args:
            path (str): Path to store the file with the metadata information.
        """
        if isinstance(self.category, str):
            category = Categories[category.upper()]
        else:
            category = self.category.value
        with open(path+'/metadata.json', 'w') as f:
            metadata = {}
            metadata["category"] = category
            metadata["name"] = self.name
            metadata["last_modified"] = self.last_modified
            metadata["description"] = self.description
            metadata["version"] = self.version
            metadata["features"] = self.features
            if int(self.previous_experimentID) > -1: metadata["previous_experimentID"] = int(self.previous_experimentID)
            f.write(json.dumps(metadata, sort_keys=True, indent=4))




service = """
apiVersion: v1
kind: Service
metadata:
  name: service-easier-training
  labels:
    app: easier-training
spec:
  type: ClusterIP
  ports:
  - name: "5100"
    port: 5100
    targetPort: 5100
  selector:
    app: easier-training
"""

service_distributed = """
apiVersion: v1
kind: Service
metadata:
  name: service-easier-dist-training
  labels:
    app: easier-dist-training
spec:
  type: ClusterIP
  ports:
  - name: "6010"
    port: 6010
    targetPort: 6010
  - name: "6006"
    port: 6006
    targetPort: 6006
  selector:
    app: easier-dist-training
"""


import time as time_lib
from datetime import datetime
import argparse
import os, sys
import joblib
import json
import tensorflow as tf
import pandas
import numpy
import joblib
import tempfile

from easierSDK.easier import EasierSDK
from easierSDK.classes.categories import Categories
from easierSDK.classes.model_metadata import ModelMetadata
from easierSDK.classes.easier_model import EasierModel
import easierSDK.classes.constants as constants
from easierSDK import datasetsAPI

# Important readings
# https://www.tensorflow.org/tutorials/distribute/input
# https://www.tensorflow.org/api_docs/python/tf/distribute/MultiWorkerMirroredStrategy
# https://www.tensorflow.org/tutorials/distribute/multi_worker_with_keras


# Variable definition
easier = None
easier_model = None
x = None
y = None
x_test = None
y_test = None

# Environment variables loading
tf_config = json.loads(os.environ['TF_CONFIG'])
num_workers = len(tf_config['cluster']['worker'])

checkpoint_dir = os.getenv('CHECKPOINT_DIR', '/train/checkpoint')
tf_log_dir = os.getenv('LOG_DIR', '/train/tf_logs/')
data_dir = os.getenv('DATA_DIR', '/train/data')
model_dir = os.getenv('MODEL_DIR', '/train/model')
model_name = os.getenv('MODEL_NAME', 'model')

per_worker_batch_size = os.getenv('WORKER_BATCH_SIZE', 64)
global_batch_size = per_worker_batch_size * num_workers
save_freq = os.getenv('SAVE_FREQ', 10)

def easier_dataset(input_context):
    # Ideally here we would call the easierSDK to get the training data.
    # Create dataset object
    global x
    global y
    dataset = tf.data.Dataset.from_tensor_slices((x, y))
    
    # Shard the dataset between the workers
    dataset = dataset.shard(input_context.num_input_pipelines, input_context.input_pipeline_id)
    
    # Repeat this dataset infinitely (every time this function is called, a counter is reduced, if we do not set this, then the dataset is returned only once)
    dataset = dataset.repeat()

    # Get batch_size from global_batch_size and batch the dataset
    batch_size = input_context.get_per_replica_batch_size(global_batch_size)
    dataset = dataset.batch(batch_size)
    
    # Prefetch 2 batches per device.
    dataset = dataset.prefetch(2) 
    
    return dataset

def compile_model(easier_model:EasierModel):
    # TF Compile should be done within distributed strategy
    easier_model.model.compile(
        optimizer=easier_model.training_metadata.optimizer, 
        loss=easier_model.training_metadata.loss,
        metrics=easier_model.training_metadata.metrics
    )
    easier_model.model.summary()
    return easier_model

if __name__ == "__main__":
    
    start = time_lib.time()

    # Initializations
    cluster_resolver = tf.distribute.cluster_resolver.TFConfigClusterResolver()

    strategy = tf.distribute.MultiWorkerMirroredStrategy(cluster_resolver=cluster_resolver)

    print("Task type, id: " + str(cluster_resolver.task_type) + ", " + str(cluster_resolver.task_id))

    # ------------------------- Distributed Parameter Training -------------------------
    # if cluster_resolver.task_type in ("worker", "ps"):
    #     # start a TensorFlow server and wait.
    #     print("Starting TF server...")
    #     server = tf.distribute.Server(
    #         cluster_resolver.cluster_spec(),
    #         job_name=cluster_resolver.task_type,
    #         task_index=cluster_resolver.task_id,
    #         protocol=None,
    #         start=True)
    #     server.join()
    # elif cluster_resolver.task_type == "evaluator":
    #     # run side-car evaluation
    #     continue
    # else:
        # run the coordinator/chief
    # -------------------------------------------------------------------------

    easier_user = os.getenv("easier_user")
    easier_password = os.getenv("easier_password")
    ignore_cert_warning = os.getenv("ignore_cert_warning")
    minio_region = os.getenv("minio_region")
    minio_secure = os.getenv("minio_secure")
    easier_url = os.getenv("easier_url")
    easier = EasierSDK(easier_user=easier_user, easier_password=easier_password, easier_url=easier_url, minio_secure=minio_secure, minio_region=minio_region, ignore_cert_warning=ignore_cert_warning)

    repo = os.getenv("repo")
    category = os.getenv("category")
    model_name = os.getenv("model_name")
    experimentID = os.getenv("experimentID")
    data_path = os.getenv("data_path")

    # Ideally, this should be called within the easier_dataset function. But we keep it here to obtain also the validation data
    x, y, x_test, y_test = easier.training.get_training_data(repo_path=data_path)

    # ------------------------- DATASET DISTRIBUTION -------------------------

    # ------------- Method 1 -------------
    # options = tf.data.Options()
    # options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA    
    # Read data from /train/data
    # dataset = tf.data.Dataset.from_tensor_slices((x, y)).with_options(options)
    # multi_worker_dataset = strategy.experimental_distribute_dataset(dataset, options)
    # ---------------------------------------

    # ------------- Method 2 - supposedly better --------
    multi_worker_dataset = strategy.distribute_datasets_from_function(easier_dataset)
    # , 
    #                     options=tf.distribute.InputOptions(
    #                         experimental_prefetch_to_device=True,
    #                         experimental_replication_mode=tf.distribute.InputReplicationMode.PER_WORKER,
    #                         experimental_place_dataset_on_device=False
    # ))
    # -----------------------------------------------

    # -------------------------------------------------------------------------

    with strategy.scope():
        easier_model = easier.models.get_model(repo_name=repo, category=category, model_name=model_name, experimentID=experimentID, load_level=constants.FULL, training=True)

        # Model building/compiling need to be within `strategy.scope()`.
        multi_worker_model = compile_model(easier_model)

    # if cluster_resolver.task_id == 0:
    #     os.system("nohup tensorboard --logdir " + tf_log_dir + " --bind_all &")
    #     tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir=tf_log_dir + datetime.now().strftime("%Y%m%d-%H%M%S"))
    #     callbacks = [tensorboard_callback]
    # else:
    #   callbacks = None
    # callbacks = [
    #         # This callback saves a SavedModel every epoch
    #         # We include the current epoch in the folder name.
    #         tf.keras.callbacks.ModelCheckpoint(
    #             filepath=checkpoint_dir + "/ckpt-{epoch}", save_freq=save_freq
    #         )
    #     ]
    callbacks = None
    # steps_per_epoch must be set when using distributed training. Usually it is set as len(x)/global_batch_size
    history = multi_worker_model.model.fit(multi_worker_dataset, validation_data=(x_test, y_test), 
                epochs=easier_model.training_metadata.epochs, verbose=2, 
                steps_per_epoch=int(len(x)/global_batch_size),
                callbacks=callbacks)
    
    # If we are chief (worker 0)
    if cluster_resolver.task_id == 0:
        
        multi_worker_model.model.save(checkpoint_dir + "/final_model.h5")
        
        easier.training.clear_training_data(repo_path=data_path, repo_name=repo)

        easier.models._finish_model_training(multi_worker_model, repo_name=repo, category=category, experimentID=experimentID)

    end = time_lib.time()
    print("Total execution time: " + str(end))

    # Successfull exit
    sys.exit()


ingress = """
apiVersion: networking.k8s.io/v1beta1
kind: Ingress
metadata:
  name: ingress-easier-serving
  kubernetes.io/ingress.allow-http: "true"
  nginx.ingress.kubernetes.io/backend-protocol: HTTP
  nginx.ingress.kubernetes.io/secure-backends: "false"
  labels:
    app: easier-serving
spec:
  tls:
  - secretName: wildcard-easier-ai.eu
  rules:
  - host: easier-ai.eu
    http:
      paths:
      - backend:
          serviceName: service-easier-serving
          servicePort: 5000
"""


#   Copyright  2020 Atos Spain SA. All rights reserved.
 
#   This file is part of EASIER AI.
 
#   EASIER AI is free software: you can redistribute it and/or modify it under the terms of Apache License, either version 2 of the License, or
#   (at your option) any later version.
 
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT ANY WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING 
#   BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT,
#   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#   WHETHER IN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
#   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#   See  LICENSE file for full license information  in the project root.

import sys
sys.path.append('..')
sys.path.append('.')

import time
import json
import os
import joblib
import random
import string
from shutil import copyfile
import tensorflow as tf
import numpy

import easierSDK.classes.constants as constants
from easierSDK.classes.categories import Categories
from easierSDK.classes.model_metadata import ModelMetadata
from easierSDK.classes.training_metadata import TrainingMetadata

class EasierModel():
    """Class to control the Easier model's objects.
    """

    metadata = None
    training_metadata = None
    model = None
    scaler = None
    label_encoder = None
    feature_encoder = None
    tf_lite_model_path = ''
    tpu_model_path = ''
    gpu_model_path = ''
    model_dir = ''
    samples = None

    def __init__(self, metadata:ModelMetadata=None):
        """Constructor for the model class.

        Args:
            metadata (ModelMetadata, optional): model's metadata. Defaults to None.
        """
        self.metadata = metadata

    def set_metadata(self, metadata:ModelMetadata):
        """Set model's metadata.

        Args:
            metadata (ModelMetadata): model's metadata.
        """
        self.metadata = metadata
    def get_metadata(self) -> ModelMetadata:
        """Return model's metadata.

        Returns:
            ModelMetadata: model's metadata.
        """
        return self.metadata

    def set_training_metadata(self, training_metadata:TrainingMetadata):
        """Set model's training metadata.

        Args:
            metadata (TrainingMetadata): model's training metadata.
        """
        self.training_metadata = training_metadata
    def get_training_metadata(self) -> TrainingMetadata:
        """Return model's training metadata.

        Returns:
            TrainingMetadata: model's training metadata.
        """
        return self.training_metadata

    def set_model(self, model:object):
        """Set the model variable.

        Args:
            model (object): Your model variable.
        """
        self.model = model
    def get_model(self) -> object:
        """Return the model variable.

        Returns:
            object: model.
        """
        return self.model
    
    def set_scaler(self, scaler:object):
        """Set scaler for the model.

        Args:
            scaler (object): scaler for the model.
        """
        self.scaler = scaler
    def get_scaler(self) -> object:
        """Return scaler for the model.

        Returns:
            object: scaler for the model.
        """
        return self.scaler
    
    def set_label_encoder(self, label_encoder:object):
        """Set label encoder for the model.

        Args:
            label_encoder (object): label encoder for the model.
        """
        self.label_encoder = label_encoder
    def get_label_encoder(self) -> object:
        """Return label encoder for the model.

        Returns:
            object: label encoder for the model.
        """
        return self.label_encoder
    
    def set_feature_encoder(self, feature_encoder:object):
        """Set feature encoder for the model.

        Args:
            feature_encoder (object): feature encoder for the model.
        """
        self.feature_encoder = feature_encoder
    def get_feature_encoder(self) -> object:
        """Return feature encoder for the model.

        Returns:
            object: feature encoder for the model.
        """
        return self.feature_encoder
    
    def set_tf_lite_model_path(self, tf_lite_model_path:str):
        """Set path to store the TFLITE version of the model.

        Args:
            tf_lite_model_path (str): path to store the TFLITE version of the model.
        """
        self.tf_lite_model_path = tf_lite_model_path
    def get_tf_lite_model_path(self):
        """Return path to TFLITE version of the model.

        Returns:
            [type]: path to TFLITE version of the model.
        """
        return self.tf_lite_model_path

    def set_tpu_model_path(self, tpu_model_path:str):
        """Set path to store the TPU version of the model.

        Args:
            tpu_model_path (str): path to store the TPU version of the model.
        """
        self.tpu_model_path = tpu_model_path
    def get_tpu_model_path(self):
        """Return path to TPU version of the model.

        Returns:
            str: path to TPU version of the model.
        """
        return self.tpu_model_path
    
    def set_gpu_model_path(self, gpu_model_path:str):
        """Set path to store the GPU version of the model.

        Args:
            gpu_model_path (str): path to store the GPU version of the model.
        """
        self.gpu_model_path = gpu_model_path
    def get_gpu_model_path(self) -> str:
        """Return path to GPU version of the model.

        Returns:
            str: path to GPU version of the model.
        """
        return self.gpu_model_path

    def set_representative_data(self, samples:list):
        """Set representative data of the model.

        Args:
            samples (list): representative data of the model.
        """
        self.samples = samples
    def get_representative_data(self) -> list:
        """Return representative data of the model.

        Returns:
            list: representative data of the model.
        """
        return self.samples
    
    def _get_random_string(self) -> str:
        """Returns a random string.

        Returns:
            str: random string
        """
        letters = string.ascii_lowercase
        return ''.join(random.choice(letters) for i in range(random.randint(1, 16))) 

    def store(self, model_path:str=None, storage_level=constants.FULL, print_files:bool=True) -> str:
        """Save the model's objects to files.

        Args:
            model_path (str, optional): Path to which to store the files. Defaults to None to create a random local folder.
            print_files (bool, optional): Whether to print which files are stored or not. Defaults to True.

        Returns:
            str: path in which the files were stored.
        """
        files_print = ""
        if model_path is None:
            path = '/tmp/' + self._get_random_string()
            while os.path.isdir(path): path += self._get_random_string()
            os.mkdir(path)
        else:
            path = model_path
        
        if self.metadata:
            self.metadata.dump_to_file(path)

        if storage_level == constants.FULL:
            try:
                self.model.save(path + "/" + self.metadata.name + '.' + constants.MODEL_EXTENSION)
                files_print += "\n\t- " + self.metadata.name + '.' + constants.MODEL_EXTENSION    
            except Exception as e:
                print("[ERROR] Couldn't save model as HDF5: " + str(e))
                print("Trying with joblib...")
                try: 
                    joblib.dump(self.model, path + "/" + 'model.h5')
                except Exception as e2:
                    print("[ERROR] Couldn't save model with joblib: " + str(e2))
        
        if storage_level == constants.FULL or storage_level == constants.WEIGHTS_ONLY:
            try:
                self.model.save_weights(path + "/" + self.metadata.name + '.' + constants.WEIGHTS_EXTENSION)
                files_print += "\n\t- " + self.metadata.name + '.' + constants.WEIGHTS_EXTENSION
            except Exception as e:
                print("[ERROR] Couldn't save weights as HDF5: " + str(e))

        try:
            json_config = self.model.to_json()
            with open(path + "/" + self.metadata.name + '.' + constants.MODEL_STRUCTURE_EXTENSION, 'w') as json_file:
                json.dump(json_config, json_file)
            files_print += "\n\t- " + self.metadata.name + '.' + constants.MODEL_STRUCTURE_EXTENSION
        except Exception as e:
            print("[ERROR] Couldn't save model structure to JSON: " + str(e))

        if self.training_metadata:
            self.training_metadata.dump_to_file(path)
        
        if self.scaler: 
            joblib.dump(self.scaler, path + "/" + 'scaler.s.pkl')
            files_print += "\n\t- scaler.s.pkl"
        if self.label_encoder: 
            joblib.dump(self.label_encoder, path + "/" + 'label_encoder.e.pkl')
            files_print += "\n\t- label_encoder.e.pkl"
        if self.feature_encoder: 
            joblib.dump(self.feature_encoder, path + "/" + 'feature_encoder.ohe.pkl')
            files_print += "\n\t- feature_encoder.ohe.pkl"
        if self.tf_lite_model_path: 
            copyfile(self.tf_lite_model_path, path + "/" + self.tf_lite_model_path.split('/')[-1])
            files_print += "\n\t- " + self.tf_lite_model_path.split('/')[-1]
        if self.tpu_model_path: 
            copyfile(self.tpu_model_path, path + "/" + self.tpu_model_path.split('/')[-1])
            files_print += "\n\t- " + self.tpu_model_path.split('/')[-1]
        if self.gpu_model_path: 
            copyfile(self.gpu_model_path, path + "/" + self.gpu_model_path.split('/')[-1])
            files_print += "\n\t- " + self.gpu_model_path.split('/')[-1]
        
        if print_files:
            print("Stored files in " + path + ":\n\t- metadata.json" + ":\n\t- " + self.metadata.name + '.h5' + files_print)
        return path

    def representative_dataset_gen(self):
        """Generator for the samples that the model was trained with.

        Yields:
            list: list with a sample that the model was trained with.
        """
        for i in range(len(self.samples)):
            data = numpy.array(self.samples[i: i + 1], dtype=numpy.float32)
            yield [data]



deployment = """
apiVersion: batch/v1
kind: Job
metadata:
  labels:
    app: easier-training
  name: job-easier-training
spec:
  backoffLimit: 0
  template:
    metadata:
      labels:
        app: easier-training
    spec:
      containers:
      - name: easier-training
        image: easierai/easier_training:1.0
        imagePullPolicy: Always
        envFrom:
          - configMapRef:
              name: easier-training
        resources:
          requests:
            memory: "512Mi"
            cpu: "0.5"
          limits:
            memory: "1024Mi"
            cpu: "1"
      restartPolicy: Never
"""

deployment_distributed = """
apiVersion: batch/v1
kind: Job
metadata:
  labels:
    app: easier-dist-training
  name: job-easier-dist-training
spec:
  backoffLimit: 0
  template:
    metadata:
      labels:
        app: easier-dist-training
    spec:
      containers:
      - name: easier-dist-training
        image: easierai/easier_distributed_training:1.0
        imagePullPolicy: Always
        envFrom:
          - configMapRef:
              name: easier-dist-training
        containerPort: 5010
        resources:
          requests:
            memory: "1024Mi"
            cpu: "1"
          limits:
            memory: "2056Mi"
            cpu: "1.5"
      restartPolicy: Never
"""

deployment_pod = """
apiVersion: v1
kind: Pod
metadata:
  labels:
    app: easier-training
  name: pod-easier-training
spec:
  replicas: 1
  selector:
    matchLabels:
      app: easier-training
  template:
    metadata:
      labels:
        app: easier-training
  containers:
  - name: easier-training
    image: easierai/easier_training:1.0
    imagePullPolicy: Always
    envFrom:
      - configMapRef:
          name: easier-training
  restartPolicy: Never
"""

deployment_pvc = """
apiVersion: v1
kind: Job
metadata:
  labels:
    app: easier-training
  name: pod-easier-training
spec:
  replicas: 1
  selector:
    matchLabels:
      app: easier-training
  template:
    metadata:
      labels:
        app: easier-training
  volumes:
    - name: easier-training
      persistentVolumeClaim:
        claimName: easier-training-claim
  containers:
  - name: easier-training
    image: easierai/easier_training:1.0
    imagePullPolicy: Always
    envFrom:
      - configMapRef:
          name: easier-training
    volumeMounts:
      - mountPath: "/train"
        name: easier-training-claim
  restartPolicy: Always
  initContainers:
    - name: easier-training-load-data
      image: easierai/easier_training_load_data:1.0
      imagePullPolicy: Always
      envFrom:
      - configMapRef:
          name: easier-training
      volumeMounts:
        - mountPath: "/train"
          name: easier-training-claim
  # imagePullSecrets:
  # - name: regcred
"""


deployment = """
apiVersion: v1
kind: Pod
metadata:
  labels:
    app: easier-serving
  name: pod-easier-serving
spec:
  replicas: 1
  selector:
    matchLabels:
      app: easier-serving
  template:
    metadata:
      labels:
        app: easier-serving
  containers:
  - name: easier-serving
    image: easierai/easier_model:1.0
    imagePullPolicy: Always
    ports:
    - containerPort: 5000
    envFrom:
      - configMapRef:
          name: easier-serving
    resources:
      requests:
        memory: "512Mi"
        cpu: "0.5"
      limits:
        memory: "1024Mi"
        cpu: "1"
  restartPolicy: Always
  # imagePullSecrets:
  # - name: regcred
"""


#   Copyright  2020 Atos Spain SA. All rights reserved.
 
#   This file is part of EASIER AI.
 
#   EASIER AI is free software: you can redistribute it and/or modify it under the terms of Apache License, either version 2 of the License, or
#   (at your option) any later version.
 
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT ANY WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING 
#   BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT,
#   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#   WHETHER IN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
#   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#   See  LICENSE file for full license information  in the project root.

import minio
from minio import Minio
import urllib3
import os
import subprocess
import tensorflow as tf
import joblib
from shutil import rmtree
import json
import tensorflow.keras.backend as K
import glob
import datetime
from tqdm import trange, tqdm
import matplotlib.pyplot as plt
import matplotlib as mpl

from easierSDK.classes.categories import Categories
from easierSDK.classes.model_metadata import ModelMetadata
from easierSDK.classes.easier_model import EasierModel
from easierSDK.classes.training_metadata import TrainingMetadata
import easierSDK.classes.constants as constants


class ModelsAPI():
    """Class to control the Models API of EasierSDK.
    """

    _GLOBAL = 'global'
    _MODELS = 'models'
    _DATASETS = 'datasets'
    _BASE_MODELS_PATH = './models'
    _MAX_FOLDER_SIZE = 2000 # Kbytes = 2M
    _MAX_FOLDER_SIZE_STR = '2MB'

    def __init__(self, minio_client, my_public_repo, my_private_repo):
        """Constructor for the ModelsAPI.

        Args:
            minio_client (Minio): Minio client object with user session initialized.
            my_public_repo (str): Name of the public bucket of the user.
            my_private_repo (str): Name of the private bucket of the user.
        """
        if not os.path.isdir(self._BASE_MODELS_PATH): os.mkdir(self._BASE_MODELS_PATH)
        self.minio_client = minio_client
        self.my_public_repo = my_public_repo
        self.my_private_repo = my_private_repo

    def _count_model_experiments(self, repo_name:str, category:str, model_name:str) -> int:
        """Count number of experiments in Minio of a model.

        Args:
            repo_name (str): Name of the Minio repository.
            category (str): Category of the model.
            model_name (str): EasierModel name in repository.

        Returns:
            int: number of experiments of the model.
        """
        if isinstance(category, Categories):
           category = category.value

        num_experiments = 0
        already_counted = []
        for item in self.minio_client.list_objects(repo_name, prefix=self._MODELS + '/' + category + '/' + model_name, recursive=True):
            if item.object_name == self._MODELS + '/' + category + '/' + model_name + "/metadata.json":
                continue
            experimentID = item.object_name.split('/')[3]
            if experimentID in already_counted:
                continue
            num_experiments += 1
            already_counted.append(experimentID)
        return num_experiments, already_counted     

    def get_model_experiments(self, repo_name:str, category:Categories, model_name:str) -> list:
        """Get a list of metadata information of a specific model experiments.

        Args:
            repo_name (str): Name of the repository that contains the model.
            category (Categories): Category that contains the model.
            model_name (str): Name of the model.
            
        Returns:
            list: list of ModelMetadata objects with the experiments information.
        """
        metadata = []
        if isinstance(category, str):
           category = Categories[category.upper()]

        if not self.minio_client.bucket_exists(repo_name):
            print('[ERROR] Repository name does not exist. Please check and try again')
            return None
        num_experiments, _ = self._count_model_experiments(repo_name, category.value, model_name)
        for experiment in range(num_experiments):
            filename = self._MODELS + '/' + category.value + '/' + model_name + '/' + str(experiment) + '/' + 'metadata.json'
            try:
                minio_obj = self.minio_client.fget_object(repo_name, filename, filename)
            except minio.error.NoSuchKey as ex:
                print('[ERROR] Wrong model information, please check: category, name and experimentID and try again.')
                return None
            
            with open(filename, 'r') as f:
                model_metadata = ModelMetadata(json.load(f))
                metadata.append(model_metadata)

            if os.path.exists(filename):
                os.remove(filename)
        return metadata

    def show_model_info(self, repo_name:str, category:Categories, model_name:str, experimentID:int=-1) -> ModelMetadata:
        """Show metadata information of a specific model and/or a specific experimentID.

        Args:
            repo_name (str): Name of the repository that contains the model.
            category (Categories): Category that contains the model.
            model_name (str): Name of the model.
            experimentID (int, optional): ExperimentID of the model to which information is going to be shown. Defaults to -1 to show contextual model information without references to any experimentID.

        Returns:
            ModelMetadata: object with the model information.
        """
        if isinstance(category, str):
           category = Categories[category.upper()]

        if not self.minio_client.bucket_exists(repo_name):
            print('[ERROR] Repository name does not exist. Please check and try again')
            return None
        if experimentID == -1:
            # Show model metadata instead of experiment metadata
            # experimentID, _ = self._count_model_experiments(repo_name, category.value, model_name) - 1        
            filename = self._MODELS + '/' + category.value + '/' + model_name + '/' + 'metadata.json'
            filename_training = self._MODELS + '/' + category.value + '/' + model_name + '/' + 'training_metadata.json'
        else:
            filename = self._MODELS + '/' + category.value + '/' + model_name + '/' + str(experimentID) + '/' + 'metadata.json'
            filename_training = self._MODELS + '/' + category.value + '/' + model_name + '/' + str(experimentID) + '/' + 'training_metadata.json'
        local_file = './models/' + filename
        try:
            minio_obj = self.minio_client.fget_object(repo_name, filename, local_file)
        except minio.error.NoSuchKey as ex:
            print('[ERROR] Wrong model information, please check: category, name and experimentID and try again.')
            return None
        
        with open(local_file, 'r') as f:
            metadata = ModelMetadata(json.load(f))
            metadata.pretty_print()

        if os.path.exists(local_file):
            os.remove(local_file)

        if self.is_training(repo_name, category, model_name, experimentID):
            print("Model is still training!")
        else:
            local_file = './models/' + filename_training
            try:
                minio_obj = self.minio_client.fget_object(repo_name, filename_training, local_file)
                with open(local_file, 'r') as f:
                    metadata = TrainingMetadata(json.load(f))
                    metadata.pretty_print()

                if os.path.exists(local_file):
                    os.remove(local_file)
            except minio.error.NoSuchKey as ex:
                pass
            except Exception as e:
                print('[ERROR] There was a problem getting the training information of the model:\n' + str(e))

        return metadata
        
    def show_model_config(self, repo_name:str, category:Categories, model_name:str, experimentID:int=-1):
        """Show model's configuration in a JSON file of a specific model and/or a specific experimentID. 

        Args:
            repo_name (str): Name of the repository that contains the model.
            category (Categories): Category that contains the model.
            model_name (str): Name of the model.
            experimentID (int, optional): ExperimentID of the model to which information is going to be shown. Defaults to -1 to show contextual model information without references to any experimentID.

        Returns:
            JSON: JSON object with the model's configuration.
        """
        if isinstance(category, str):
           category = Categories[category.upper()]

        if not self.minio_client.bucket_exists(repo_name):
            print('[ERROR] Repository name does not exist. Please check and try again')
            return None
        if experimentID == -1:
            # Show last experiment ID model configuration
            experimentID, _ = self._count_model_experiments(repo_name, category.value, model_name) - 1        
    
        filename = self._MODELS + '/' + category.value + '/' + model_name + '/' + str(experimentID) + '/' + model_name + '.' + constants.MODEL_STRUCTURE_EXTENSION
        local_file = './models/' + filename
        try:
            minio_obj = self.minio_client.fget_object(repo_name, filename, local_file)
        except minio.error.NoSuchKey as ex:
            print('[ERROR] The model you indicated does not have a JSON configuration file or you provided wrong model information, please check: category, name and experimentID and try again.')
            return None
        
        with open(local_file, 'r') as f:
            configuration = json.load(f)
            print(configuration)

        return configuration

    def download(self, repo_name:str, category:Categories, model_name:str, path_to_download:str, experimentID:int=-1) -> str:
        """Downloads a model and its attached objets in a specific path.

        Args:
            repo_name (str): Name of the repository that contains the model.
            category (Categories): Category that contains the model.
            model_name (str): Name of the model.
            path_to_download (str): Local path in which to store all files.
            experimentID (int, optional): ExperimentID of the model to download. Defaults to -1 to download the last experiment.

        Returns:
            str: path under path_to_download folder where files were stored
        """
        if isinstance(category, str):
           category = Categories[category.upper()]

        if not self.minio_client.bucket_exists(repo_name):
            print('[ERROR] Repository name does not exist. Please check and try again')
            return None
        
        if experimentID == -1:
            # Download last experiment
            experimentID, _ = self._count_model_experiments(repo_name, category.value, model_name)
            experimentID -= 1
        
        if self.is_training(repo_name, category, model_name, experimentID):
            print("[ERROR] Model is still training and cannot be downloaded")
            return None

        # TODO check if models will be saved in HDF5 .h5 or tf SavedModel format (folder)
        minio_path = self._MODELS + '/' + category.value + '/' + model_name + '/' + str(experimentID)
        local_path = path_to_download + "/" + minio_path
        try:
            objects = self.minio_client.list_objects(repo_name, prefix=minio_path, recursive=True)
            with tqdm(total=len(list(objects)), colour="red", desc="Downloading model %s..." % model_name) as p:          
                for minio_object in self.minio_client.list_objects(repo_name, prefix=minio_path, recursive=True):
                    minio_obj = self.minio_client.fget_object(repo_name, minio_object.object_name, path_to_download + "/" + minio_object.object_name)
                    p.set_postfix({'file': minio_object.object_name})
                    p.update()
        except minio.error.NoSuchKey as ex:
            print('[ERROR] Wrong model name or category, please check and try again.')
            return None
        return local_path

    def _load_model_from_path(self, path:str, load_level=constants.FULL, print_files=False, easier_model:EasierModel=None, model_name=None) -> EasierModel:
        if easier_model is None: easier_model = EasierModel()

        if model_name is None:
            if easier_model.metadata is None or easier_model.metadata.name is None:
                try:
                    model_name = glob.glob(path + '/' + '*.s.json')[0]
                    model_name = model_name.split('.s.json')[0]
                    model_name = model_name.split('/')[-1]
                except Exception:
                    try:
                        model_name = glob.glob(path + '/' + '*.h5')[0]
                        model_name = model_name.split('.h5')[0]
                        model_name = model_name.split('/')[-1]
                    except Exception as e:
                        print("[ERROR] Could not find a mathing file for the model.")
            else:
                model_name = easier_model.metadata.name

        if load_level == constants.CONFIG_ONLY or load_level == constants.WEIGHTS_ONLY:
            # First we need to load the config/structure of the model as json
            json_model = None
            try:
                with open(path + '/' + model_name + '.s.json') as json_file:
                    json_model = json.load(json_file)
                    json_model = tf.keras.models.model_from_json(json_model)
                if print_files: print("Loaded model's config from JSON. File stored in: " + path + '/' + model_name + '.s.json')
            except FileNotFoundError:
                print("The loaded model has no JSON configuration.")
            except Exception as e:
                print("[ERROR] Could not load model's config from JSON: " + str(e)) 
            
            if load_level == constants.WEIGHTS_ONLY:
                if json_model and easier_model is not None and easier_model.get_model() is None:
                    # Load the weights in case it is necessary
                    try:
                        json_model.load_weights(filepath=path + '/' + model_name + '.w.' + constants.MODEL_EXTENSION)
                    except Exception as e:
                        print("[ERROR] Could not load model's weights properly: " + str(e)) 
                else:
                    if easier_model is None:
                        print("[ERROR] Cannot load model's weights. Please, load the easier_model with an appropiate model architecture. ") 
                    elif easier_model.get_model() is not None:
                        try:
                            easier_model.model.load_weights(path + '/' + model_name + '.w.' + constants.MODEL_EXTENSION)
                        except Exception as e:
                            print("[ERROR] Could not load model's weights properly: " + str(e)) 
                    else:
                        print("[ERROR] Cannot load model's weights. Please, load the easier_model with an appropiate model architecture. ") 
            # Set the model for EasierModel variable
            easier_model.set_model(json_model)
        elif load_level == constants.FULL:
            try:
                # Load model from file (HDF5 default)
                easier_model.set_model(tf.keras.models.load_model(path + '/' + model_name + '.' + constants.MODEL_EXTENSION))
                if print_files: print("Loaded model. File stored in: " + path + '/' + model_name + '.' + constants.MODEL_EXTENSION)
            except Exception as e:
                print("[ERROR] Could not load full model: " + str(e))
        else:
            print("[ERROR] Load level: " + str(load_level) + " not supported. Use constants.CONFIG_ONLY, constants.WEIGHTS_ONLY or constants.FULL")

        return easier_model


    def _load_file(self, path:str, load_level=constants.FULL, print_files=False, easier_model:EasierModel=None, model_name=None) -> EasierModel:
        """Loads files from a path onto an EASIER model variable depending on the file's extension.

        Args:
            path (str): path of the file
            print_files (bool, optional): Whether or not to print which files are being loaded. Defaults to False.
            easier_model (EasierModel, optional): EASIER EasierModel variable to to load the files. Defaults to None to create a new model.

        Returns:
            EasierModel: EASIER model variable with file loaded
        """
        if easier_model is None: easier_model = EasierModel()

        easier_model = self._load_model_from_path(path, load_level=load_level, print_files=print_files, easier_model=easier_model, model_name=model_name)
        objects = os.listdir(path)
        for obj in objects:
            extension = obj.split(".")[-1]

            if extension == constants.MODEL_EXTENSION:
               continue          

            elif extension == constants.PICKLE_EXTENSION:
                extension_pkl = obj.split(".")[-2:]
                extension_pkl = '.'.join(extension_pkl)
                if extension_pkl == constants.SCALER_EXTENSION:
                    easier_model.set_scaler(joblib.load(path + '/' + obj))
                    if print_files: print("Loaded scaler. File stored in: " + path + '/' + obj)
                elif extension_pkl == constants.LABELENCODER_EXTENSION:
                    easier_model.set_label_encoder(joblib.load(path + '/' + obj))
                    if print_files: print("Loaded label encoder. File stored in: " + path + '/' + obj)
                elif extension_pkl == constants.ONEHOTENCODER_EXTENSION:
                    easier_model.set_feature_encoder(joblib.load(path + '/' + obj))
                    if print_files: print("Loaded one hot encoder. File stored in: " + path + '/' + obj)

            elif extension == constants.JSON_EXTENSION:
                obj_name = obj.split('.')[-2]
                if obj_name == 'metadata':
                    with open(path + '/' + obj, 'r') as f:
                        metadata = ModelMetadata(f=json.load(f))
                        easier_model.set_metadata(metadata)
                    if print_files: print("Loaded model's metadata. File stored in: " + path + '/' + obj)
                elif obj_name == 'training_metadata':
                    with open(path + '/' + obj, 'r') as f:
                        metadata = TrainingMetadata(f=json.load(f))
                        easier_model.set_training_metadata(metadata)
                    if print_files: print("Loaded model's metadata. File stored in: " + path + '/' + obj)
                else:
                    # Load model from JSON configuration
                    continue
                        
            elif extension == constants.TF_LITE_EXTENSION: 
                extension_tflite = obj.split("_")[-1]    
                if '_' + extension_tflite == constants.TPU_EXTENSION:
                    easier_model.set_tpu_model_path(path + '/' + obj)
                    if print_files: print("Downloaded tpu model version. File stored in: " + path + '/' + obj)
                else:
                    easier_model.set_tf_lite_model_path(path + '/' + obj)
                    if print_files: print("Downloaded tf lite model version. File stored in: " + path + '/' + obj)
            else:
                if print_files: print("Couldn't load file with extension: ." + str(extension))

        return easier_model

    def load_from_local(self, path:str, load_level=constants.FULL, print_files=False) -> EasierModel:
        """Loads a model into memory from a local path.

        Args:
            path (str): Local path where model files are stored.
            print_files (bool, optional): Whether to print which files are loaded. Defaults to False.

        Returns:
            EasierModel: object with the loaded model and its attached elements like scalers.
        """
        easier_model = self._load_file(path, load_level=load_level, print_files=print_files)
                
        return easier_model

    def is_training(self, repo_name:str, category:Categories, model_name:str, experimentID:int=-1):
        """Checks if a model is in training state.

        Args:
            repo_name (str): Name of the repository that contains the model.
            category (Categories): Category that contains the model.
            model_name (str): Name of the model.
            experimentID (int, optional): ExperimentID of the model to download. Defaults to -1 to load the last experiment.
        """
        if isinstance(category, str):
           category = Categories[category.upper()]

        if not self.minio_client.bucket_exists(repo_name):
            print('[ERROR] Repository name does not exist. Please check and try again')
            return True
        
        if experimentID == -1:
            # Check last experiment
            experimentID, _ = self._count_model_experiments(repo_name, category.value, model_name)
            experimentID -= 1
        
        minio_path = self._MODELS + '/' + category.value + '/' + model_name + '/' + str(experimentID)
        
        try:
            objects = self.minio_client.list_objects(repo_name, prefix=minio_path, recursive=True)
            for minio_object in objects:
                if str(minio_object.object_name).split('/')[-1] == '.training':
                    return True
        except minio.error.NoSuchKey as ex:
            print('[ERROR] Wrong model name or category, please check and try again.')
            return True
        return False

    def _finish_model_training(self, easier_model:EasierModel, repo_name:str, category:Categories, experimentID:int):
        """Uploads a trained model into the repository and removes its training state.

        Args:
            easier_model (EasierModel): Mode to be uploaded.
            repo_name (str): Name of the repository that contains the model.
            category (Categories): Category that contains the model.
            experimentID (int, optional): ExperimentID of the model to download. Defaults to -1 to load the last experiment.
        """
        if easier_model.metadata is None or easier_model.metadata.name is None:
            print("[ERROR] Cannot upload model without a ModelMetadata and a ModelMetadata.name")
            return False
        
        if '.' in easier_model.metadata.name:
            print("[ERROR] Cannot upload model with '.' in its name. Please modify the name accordingly.")
            return False

        if not category:
            if isinstance(easier_model.metadata.category, Categories):
                category = easier_model.metadata.category.value
            else:
                category = easier_model.metadata.category
        else:
            if isinstance(category, Categories):
                category = category.value

        # Store files to be uploaded
        path = easier_model.store(storage_level=constants.FULL, print_files=False)
        minio_path = self._MODELS + '/' + category + '/' + easier_model.metadata.name
                                 
        try:
            for obj in os.listdir(path):
                self.minio_client.fput_object(repo_name, minio_path + '/' + str(experimentID) + '/' + obj, path + '/' + obj)
            # If it is a new model, upload a metadata file on top of experiments representing the first model.
            if experimentID == 0:
                self.minio_client.fput_object(repo_name, minio_path + '/' + "metadata.json", path + '/' + "metadata.json")
        except Exception as ex:
            print("[ERROR] Error uploading model to MINIO: " + str(ex))
            return False
        
        try:
            self.minio_client.remove_object(repo_name, minio_path + '/' + str(experimentID) + '/' + ".training")
        except Exception as ex:
            print("[ERROR] Error deleting training state of model in MINIO: " + str(ex))
            return False

        return True

    def get_model(self, repo_name:str, category:Categories, model_name:str, experimentID:int=-1, load_level=constants.FULL, print_files=False, training=False) -> EasierModel:
        """Loads a model into memory from a repository.

        Args:
            repo_name (str): Name of the repository that contains the model.
            category (Categories): Category that contains the model.
            model_name (str): Name of the model.
            experimentID (int, optional): ExperimentID of the model to download. Defaults to -1 to load the last experiment.
            load_level (str, optional): Levels of loading a model: FULL_MODEL, WEIGHTS_ONLY, CONFIG_ONLY. Defaults to FULL_MODEL.
            print_files (bool, optional): Whether to print which files are loaded. Defaults to False.

        Returns:
            EasierModel:  object with the loaded model and its attached elements like scalers.
        """
        if isinstance(category, str):
           category = Categories[category.upper()]

        if not training:
            if self.is_training(repo_name, category, model_name, experimentID):
                print("[ERROR] Model is still training and cannot be downloaded")
                return None

        if not os.path.isdir('/tmp/download'): os.mkdir('/tmp/download')

        if experimentID == -1:
            # Load last experiment
            experimentID, _ = self._count_model_experiments(repo_name, category.value, model_name)
            experimentID -= 1
        
        path = self.download(repo_name, category, model_name, path_to_download='/tmp/download', experimentID=experimentID)

        if print_files: print("Files are stored in: " + path)
        
        easier_model = self._load_file(path, load_level=load_level, print_files=print_files, model_name=model_name)
        easier_model.metadata.previous_experimentID = experimentID
        
        return easier_model

    def upload(self, easier_model:EasierModel, category:Categories=None, public:bool=False, storage_level:str=constants.FULL, remove_dir:bool=True, training:bool=False) -> bool:
        """Upload a model and its attached elements to the user's repository. *Important*: This function updates your model version and last_experimentID according to repository's information. 

        Args:
            category (Categories): Category that contains the model.
            easier_model (EasierModel): Object with the model and its attached elements.
            public (bool, optional): Whether to upload the model in the public version of the repository. Defaults to False.
            storage_level (str, optional): Levels of storage for a model: FULL_MODEL, WEIGHTS_ONLY, CONFIG_ONLY
            remove_dir (bool, optional): Whether to remove the folder in which the files to be uploaded were created. Defaults to True.
            training (bool, optional): Internal parameter to check whether the model is uploaded for training: **IMPORTANT** Setting this to True will prevent the model to be downloaded.

        Returns:
            bool: True if successful upload. False otherwise.
        """
        if easier_model.metadata is None or easier_model.metadata.name is None:
            print("[ERROR] Cannot upload model without a ModelMetadata and a ModelMetadata.name")
            return False
        
        if '.' in easier_model.metadata.name:
            print("[ERROR] Cannot upload model with '.' in its name. Please modify the name accordingly.")
            return False

        if storage_level.lower() != constants.FULL and storage_level.lower() != constants.WEIGHTS_ONLY and storage_level.lower() != constants.CONFIG_ONLY:
            print("[ERROR] Model storage level (" + str(storage_level) + ") not supported.")
            print("Supported model storage levels: full (constants.FULL), weights_only (constants.WEIGHTS_ONLY), config_only (constants.CONFIG_ONLY)")
            return False

        if not category:
            if isinstance(easier_model.metadata.category, Categories):
                category = easier_model.metadata.category.value
            else:
                category = easier_model.metadata.category
        else:
            if isinstance(category, Categories):
                category = category.value

        if public:
            bucket = self.my_public_repo
        else:
            bucket = self.my_private_repo
        
        # Create bucket if doesn't exist
        if not self.minio_client.bucket_exists(bucket): self.minio_client.make_bucket(bucket, location='es')

        # Check last experiment
        last_experimentID, _ = self._count_model_experiments(bucket, category, easier_model.metadata.name)

        # Update version to new experimentID
        easier_model.metadata.previous_experimentID = easier_model.metadata.version
        easier_model.metadata.version = last_experimentID

        # Update last modified timestamp in metadata 
        easier_model.metadata.last_modified = datetime.datetime.now().strftime("%Y/%m/%d %H:%M:%S")
        
        # Store files to be uploaded
        path = easier_model.store(storage_level=storage_level, print_files=False)
        minio_path = self._MODELS + '/' + category + '/' + easier_model.metadata.name

        # If model is uploaded to be trained, then a file will indicate that it is training and not ready to be downloaded
        if training:
            with open(path+'/.training', 'w') as f:
                f.write('')
                                 
        try:
            files = tqdm(os.listdir(path), colour="green", desc="Uploading %s..." % easier_model.metadata.name)
            for obj in files:
                self.minio_client.fput_object(bucket, minio_path + '/' + str(last_experimentID) + '/' + obj, path + '/' + obj)
                files.set_postfix({'file': obj})
            # If it is a new model, upload a metadata file on top of experiments representing the first model.
            if last_experimentID == 0:
                self.minio_client.fput_object(bucket, minio_path + '/' + "metadata.json", path + '/' + "metadata.json")
            if remove_dir: rmtree(path)
        except Exception as ex:
            print("[ERROR] Error uploading model to MINIO: " + str(ex))
            return False
        print("Uploaded model: \n")
        easier_model.metadata.pretty_print()
        return True

    def plot_metrics_history(self, history, metrics_names):
        mpl.rcParams['figure.figsize'] = (25,25)
        colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        plt.figure()
        for n, metric in enumerate(metrics_names):
            name = metric.replace("_"," ").capitalize()
            plt.subplot(int(len(metrics_names)/2),int(len(metrics_names)/2),n+1)
            plt.plot(history.epoch, history.history[metric], color=colors[0], label='Train')
            plt.plot(history.epoch, history.history['val_'+metric],
                    color=colors[0], linestyle="--", label='Val')
            plt.xlabel('Epoch')
            plt.ylabel(name)
            if metric == 'loss':
                plt.ylim([0, plt.ylim()[1]])
            elif metric == 'auc':
                plt.ylim([0.8,1])
            else:
                plt.ylim([0,1])

            plt.legend()

    def compile_tflite(self, easier_model:EasierModel, calibration_data: list, path:str=_BASE_MODELS_PATH + "/storage/"):
        """Compiles a Tensorflow model to its Lite version.

        Args:
            easier_model (EasierModel): EasierModel object with the Tensorflow model to be compiled.
            calibration_data (list): List with some example data that the model was trained with.
            path (str, optional): Path to store the tflite file. Defaults to "./models/storage/".
        """
        # TODO check model type (h5 vs SavedModel)
        # TODO check usage of path
        if not os.path.isdir(path): os.mkdir(path)
        if not os.path.isdir("/tmp/storage/"): os.mkdir("/tmp/storage/")
        easier_model.set_representative_data(samples = calibration_data)

        if easier_model.model is None and easier_model.model_dir != '':
            converter = tf.lite.TFLiteConverter.from_saved_model(easier_model.model_dir)
        elif easier_model.model is not None:
            if tf.__version__.split(".")[0] == '1':
                easier_model.model.save("/tmp/storage/" + easier_model.metadata.name + "." + constants.MODEL_EXTENSION)
                # Clear graph in prep for next step.
                try:
                    K.clear_session()
                except Exception as e:
                    pass                
                converter = tf.lite.TFLiteConverter.from_keras_model_file("/tmp/storage/" + easier_model.metadata.name + "." + constants.MODEL_EXTENSION)
            else:    
                converter = tf.lite.TFLiteConverter.from_keras_model(easier_model.model)

        converter.representative_dataset = easier_model.representative_dataset_gen
        converter.optimizations = [tf.lite.Optimize.DEFAULT]

        try:
            tflite_model = converter.convert()
        except Exception as e:
            print("Error converting model to tf lite: " + str(e))
            return

        if easier_model.model_dir != '':
            tf_lite_model_path = path + easier_model.model_dir + "/" + easier_model.metadata.name + "." + constants.TF_LITE_EXTENSION
            open(tf_lite_model_path, "wb").write(tflite_model)
            print("Converted tf model " + str(easier_model.model_dir) + " to tf lite")
        else:
            tf_lite_model_path = path + easier_model.metadata.name + "." + constants.TF_LITE_EXTENSION
            open(tf_lite_model_path, "wb").write(tflite_model)
            # Clear graph in prep for next step.
            try:
                K.clear_session()
            except Exception as e:
                pass
            print("Converted keras model " + easier_model.metadata.name + " to tf lite")

        easier_model.set_tf_lite_model_path(tf_lite_model_path)

    def compile_tpu(self,  easier_model:EasierModel, calibration_data, path:str=_BASE_MODELS_PATH + "/storage/"):
        """Compiles a Tensorflow model to its TPU version.

        Args:
            easier_model (EasierModel): EasierModel object with the Tensorflow model to be compiled.
            calibration_data (list): List with some example data that the model was trained with.
            path (str, optional): Path to store the tflite.edge_tpu file. Defaults to "./models/storage/".
        """
        # TODO check model type (h5 vs SavedModel)
        # TODO check usage of path

        if not os.path.isdir(path): os.mkdir(path)
        if not os.path.isdir("/tmp/storage/"): os.mkdir("/tmp/storage/")
        easier_model.set_representative_data(samples = calibration_data)

        if easier_model.model_dir != '':
            converter = tf.lite.TFLiteConverter.from_saved_model(easier_model.model_dir)
        else:
            if tf.__version__.split(".")[0] == '1':
                if not os.path.isfile("/tmp/storage/" + easier_model.metadata.name + "." + constants.MODEL_EXTENSION):
                    easier_model.model.save("/tmp/storage/" + easier_model.metadata.name + "." + constants.MODEL_EXTENSION)
                try:
                    K.clear_session()
                except Exception as e:
                    pass
                converter = tf.lite.TFLiteConverter.from_keras_model_file("/tmp/storage/" + easier_model.metadata.name + "." + constants.MODEL_EXTENSION)
            else:    
                converter = tf.lite.TFLiteConverter.from_keras_model(easier_model.model)

        if tf.__version__.split(".")[0] == '1':
            converter.target_ops = [tf.lite.OpsSet.TFLITE_BUILTINS_INT8]        
        else:
            converter.target_spec.supported_ops = [tf.lite.OpsSet.TFLITE_BUILTINS_INT8]
        converter.representative_dataset = easier_model.representative_dataset_gen
        converter.inference_input_type = tf.uint8
        converter.inference_output_type = tf.uint8
        converter.optimizations = [tf.lite.Optimize.DEFAULT]

        try:
            tflite_model = converter.convert()
        except Exception as e:
            print("Error converting model to tf lite: " + str(e))
            return

        if easier_model.model_dir is not None:
            open('/tmp/storage/' + easier_model.model_dir + "/" + easier_model.metadata.name + "." + constants.TF_LITE_EXTENSION, "wb").write(
                tflite_model)
            try:
                K.clear_session()
            except Exception as e:
                pass
            print("Converted tf model " + str(easier_model.model_dir) + " to tf lite specific for TPU")
            tpu_model_path = "/tmp/storage/" + easier_model.model_dir + "/" + easier_model.metadata.name + "." + constants.TF_LITE_EXTENSION
            cmd = ['edgetpu_compiler', tpu_model_path, '-o', '/tmp/storage']
            tpu_model_path = "/tmp/storage/" + easier_model.model_dir + "/" + easier_model.metadata.name + "_edgetpu." + constants.TF_LITE_EXTENSION
        else:
            try:
                open('/tmp/storage/' + easier_model.metadata.name + "." + constants.TF_LITE_EXTENSION, "wb").write(
                    tflite_model)
                try:
                    K.clear_session()
                except Exception as e:
                    pass
                print("Converted keras model " + easier_model.metadata.name + " to tf lite specific for TPU")
            except Exception as e:
                print("Error saving tf lite model to file: " + str(e))
                try:
                    K.clear_session()
                except Exception as e:
                    pass
                return
            tpu_model_path = "/tmp/storage/" +  easier_model.metadata.name + "." + constants.TF_LITE_EXTENSION
            cmd = ['edgetpu_compiler', '-o', '/tmp/storage', tpu_model_path]
            tpu_model_path = "/tmp/storage/" +  easier_model.metadata.name + "_edgetpu." + constants.TF_LITE_EXTENSION
        try:
            res = subprocess.run(cmd, stdout=subprocess.PIPE)
            easier_model.set_tpu_model_path(tpu_model_path)
        except FileNotFoundError as e:
            print('The edge tpu complier is not installed: ' + str(e))
            return
        except Exception as e:
            print('The edge tpu complier throwed an error: ' + str(e))
            return

    # TODO check if necessary
    def _compile_gpu(self):
        pass      
    
    def _dockerize(self, easier_model:EasierModel, image_name:str=None):
        raise NotImplementedError()
    
        if not os.path.isdir("/dockerized_model"): os.mkdir("/dockerized_model")

        easier_model.store(model_path="/dockerized_model", print_files=False)
        if image_name is None: image_name = easier_model.metadata.name + '_' + easier_model._get_random_string()
        cmd = ['docker', 'build', '-f', 'Dockerfile_model', '-t', image_name]
        subprocess.run(cmd, stdout=subprocess.PIPE)

        cmd = ['docker', 'save', image_name, '--output', image_name + '.tar']
        subprocess.run(cmd, stdout=subprocess.PIPE)